# $Id: proxy.rb,v 1.1 2008/05/04 06:42:47 mooffie Exp $

# @file
# Contains the driver for our Proxy framework.

require 'socket'

#
# Extend TCPSocket with some convenience methods
#
class TCPSocket
 
  # Read HTTP's request/response headers.
  def read_headers
    headers = []
    while line = self.readline
      name, value = line.chomp.split(/:\s*/, 2)
      headers << {
        :source => line,
        :name => (name.downcase rescue name),
        :value => value
      }
      break if line.chomp.empty?
    end
    # install a method for easily getting a certain header's value:
    def headers.get name, translate=nil
      value = (find { |header| header[:name] == name } || {})[:value]
      (translate && value) ? value.send(translate) : value
    end
    return headers
  end

  # a-la PHP's passthru(): dumps data from this socket to another. 
  def pass_thru(dest, total=nil)
    if total == nil
      # pass all data
      while data = self.read(4096)
        dest.write data
      end
    else
      # pass up to 'total' bytes
      while !total.zero? and data = self.read([4096, total].min)
        dest.write data
        total -= data.size
      end
    end
  end
end


module Proxy

  #
  # 'Handler' is the base class for any handler. Handlers handle a document:
  # they can modify its headers and/or body.
  #
  class Handler

    # Whether this handler would like to handle this document
    #
    # Note: this is a class method. If positive result returned, an
    # object is instantiated.
    def self.wants_to_handle? request_headers
      false
    end

    # Whether this handler would like to handle the body
    #
    # Note: unlike 'wants_to_handle?', this method is an _object_ method. So
    # make sure 'want_to_handle?' returns 'true' as well.
    def wants_to_handle_body? response_headers
      false
    end

    # The handler is given a chance to modify the request headers.
    def handle_request_headers request_headers
    end
    
    # The handler is given a chance to modify the response headers.
    def handle_response_headers response_headers
    end

    # The handler is given a chance to modify the body
    def handle_body input, output
    end
    
    # Most handlers would want to know the content-type, so we conveniently
    # make it available for them, in this attribute.
    attr_accessor :content_type
  end

  #
  # The TCP proxy itself.
  #
  class Server
    
    attr_accessor :available_handlers

    def initialize options
      @available_handlers = options[:available_handlers] || [ ]
      @isp_proxy_addr = options[:isp_proxy_addr] or raise 'No ISP proxy address given.'
      @isp_proxy_port = options[:isp_proxy_port] or raise 'No ISP proxy port given.'
    end

    #
    # Serves requests. Done by infinite loop, so this method doesn't return.
    #
    def run
      server = TCPServer.open(8080)
      while true
        client = server.accept 
        # We're doing forks. I've tried threads but it isn't robust enough.
        if fork == nil
          # child
          session client
          exit!
        else
          # parent
          client.close
        end
      end
    end

    #
    # Handles a client session.
    #
    def session client
      remote = TCPSocket.open @isp_proxy_addr, @isp_proxy_port

      #
      # Step 1: handle the request
      #

      request_headers = client.read_headers

      handlers = []
      @available_handlers.each do |handler|
        handlers << handler.new if handler.wants_to_handle? request_headers
      end
      
      # debugging
      puts 'Handling: ' + request_headers.first[:source].chomp +
           ' (By: ' + handlers.map { |h| h.class }.join(', ') + ')'
      
      handlers.each do |handler|
        handler.handle_request_headers request_headers
      end

      # send the headers
      remote.write request_headers.map {|h| h[:source]}.join
      
      if content_length = request_headers.get('content-length', :to_i)
        # send a posted form
        client.pass_thru remote, content_length
      end
      
      #
      # Step 2: handle the response
      #

      response_headers = remote.read_headers

      content_type = response_headers.get('content-type', :downcase)

      body_handlers = handlers.select do |handler|
        handler.content_type = content_type
        handler.wants_to_handle_body? response_headers
      end

      handlers.each do |handler|
        handler.handle_response_headers response_headers
      end
     
      # receive the headers
      client.write response_headers.map {|h| h[:source]}.join

      # receive the body
      if not body_handlers.empty? and not remote.eof?
        # Ideally, we should support several handlers handling a body,
        # For the time being, only the first interested handler can handle the body.
        body_handlers.first.handle_body remote, client
      else
        remote.pass_thru client
      end
      
      client.close
      remote.close
    end

  end # class Server

end # module Proxy
